﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;

namespace Airfoil
{
    enum DataType
    {
        TRAIN, SELECT, TEST
    }

    class Program
    {
        private static List<double[]> _inputs = new List<double[]>();
        private static List<double> _target = new List<double>();
        private static List<DataType> _dt = new List<DataType>();
        private static List<U> _ulist = new List<U>();
        private static U _bigU = new U();
        private static Random _rnd = new Random();
        private static double _targetMin;
        private static double _targetMax;

        static double[] GetRandomSignal(int size, double min, double max)
        {
            double[] x = new double[size];
            for (int i = 0; i < x.Length; ++i)
            {
                x[i] = _rnd.Next() % 100;
            }
            double xmin = x.Min();
            double xmax = x.Max();
            for (int i = 0; i < x.Length; ++i)
            {
                x[i] = (x[i] - xmin) / (xmax - xmin) * (max - min) + min;
            }
            return x;
        }

        static void GenerateInitialOperators(int M, int nx)
        {
            _ulist.Clear();
            //find min, max
            int size = _inputs[0].Length;
            double[] min = new double[size];
            double[] max = new double[size];
            for (int k = 0; k < size; ++k)
            {
                min[k] = _inputs[0][k];
                max[k] = _inputs[0][k];
            }
            foreach (double[] data in _inputs)
            {
                for (int k = 0; k < size; ++k)
                {
                    if (data[k] < min[k]) min[k] = data[k];
                    if (data[k] > max[k]) max[k] = data[k];
                }
            }
            /////// end

            for (int counter = 0; counter < M; ++counter)
            {
                double[] z = GetRandomSignal(_target.Count, _targetMin, _targetMax); 

                U u = new U();
                u.Initialize(_inputs[0].Length, nx, min, max);
                for (int count = 1; count < 10; ++count)
                {
                    double accuracy = 0.0;
                    int cnt = 0;
                    for (int i = 0; i < _inputs.Count; ++i)
                    {
                        if (DataType.TRAIN != _dt[i]) continue;
                        double[] data = _inputs[i];
                        double diff = z[i] - u.GetU(data);
                        u.Update(diff, data, 0.05);
                        accuracy += diff * diff;
                        ++cnt;
                    }
                    accuracy /= cnt;
                    accuracy = Math.Sqrt(accuracy);
                    accuracy /= (_targetMax - _targetMin);
                    //Console.WriteLine("Error for initial operator {0}, {1:0.0000}", counter, accuracy); 
                }
                _ulist.Add(u);
            }
        }

        static double[] GetVector(double[] data)
        {
            int size = _ulist.Count;
            double[] vector = new double[size];
            for (int i = 0; i < size; ++i)
            {
                vector[i] = _ulist[i].GetU(data);
            }
            return vector;
        }

        static void BuildRepresentation(int M, int nx)
        {
            double[] min = new double[M];
            double[] max = new double[M];
            for (int i = 0; i < M; ++i)
            {
                min[i] = _targetMin;
                max[i] = _targetMax;
            }
            _bigU.Clear();
            _bigU.Initialize(M, nx, min, max);
            int N = _inputs.Count;
            double error = (_targetMax - _targetMin) / 150.0;
            double training_error = 0.0;

            for (int step = 0; step < 400; ++step)
            {
                double RMSE = 0.0;
                int cnt = 0;
                for (int i = 0; i < N; ++i)
                {
                    if (DataType.TRAIN != _dt[i]) continue;
                    double[] data = _inputs[i];
                    double[] v = GetVector(data);
                    double diff = _target[i] - _bigU.GetU(v);

                    if (Math.Abs(diff) > error)
                    {
                        for (int k = 0; k < _ulist.Count; ++k)
                        {
                            double derrivative = _bigU.GetDerrivative(k, v[k]);
                            if (Math.Abs(derrivative) < 0.1)
                            {
                                if (derrivative < 0.0) derrivative = -0.1;
                                else derrivative = 0.1;
                            }
                            _ulist[k].Update(diff * derrivative, data, 0.005);
                        }
                        v = GetVector(data);
                        diff = _target[i] - _bigU.GetU(v);
                        _bigU.Update(diff, v, 0.005);
                    }
                    RMSE += diff * diff;
                    ++cnt;
                }

                RMSE /= cnt;
                RMSE = Math.Sqrt(RMSE);
                RMSE /= (_targetMax - _targetMin);
                //Console.WriteLine("Error in big operator {0:0.0000}", accuracy);
                training_error = RMSE;
            }
            //Console.WriteLine("Training is completed, relative error {0:0.0000}", training_error);
        }

        static void MakeTrainSelectValidateSets()
        {
            _dt.Clear();
            foreach (double t in _target)
            {
                if (_rnd.Next() % 100 <= 60) _dt.Add(DataType.TRAIN);
                else
                {
                    if (_rnd.Next() % 100 <= 50) _dt.Add(DataType.SELECT);
                    else _dt.Add(DataType.TEST);
                }
            }
        }

        static double RMSESelection()
        {
            double RMSE = 0.0;
            int cnt = 0;
            int N = _inputs.Count;
            for (int i = 0; i < N; ++i)
            {
                if (DataType.SELECT != _dt[i]) continue;
                double[] data = _inputs[i];
                double[] v = GetVector(data);
                double diff = _target[i] - _bigU.GetU(v);
                RMSE += diff * diff;
                ++cnt;
            }
            RMSE /= cnt;
            RMSE = Math.Sqrt(RMSE);
            RMSE /= (_targetMax - _targetMin);
            return RMSE;
        }

        static double ComputeCorrelationCoeff()
        {
            int N = _inputs.Count;
            double[] targetEstimate = new double[N];
            int count = 0;
            for (int i = 0; i < N; ++i)
            {
                double[] data = _inputs[i];
                double[] v = GetVector(data);
                targetEstimate[i] = _bigU.GetU(v);
                if (DataType.TEST == _dt[i]) ++count;
            }
            double[] x = new double[count];
            double[] y = new double[count];
            count = 0;
            for (int i = 0; i < N; ++i)
            {
                if (DataType.TEST == _dt[i])
                {
                    x[count] = targetEstimate[i];
                    y[count] = _target[i];
                    ++count;
                }
            }
            return PearsonCorrelation(x, y);
        }

        static double PearsonCorrelation(double[] x, double[] y)
        {
            double xy = 0.0;
            double x2 = 0.0;
            double y2 = 0.0;
            for (int i = 0; i < x.Length; ++i)
            {
                xy += x[i] * y[i];
                x2 += x[i] * x[i];
                y2 += y[i] * y[i];
            }
            xy /= (double)(x.Length);
            x2 /= (double)(x.Length);
            y2 /= (double)(y.Length);
            double xav = x.Average();
            double yav = y.Average();
            double ro = xy - xav * yav;
            ro /= Math.Sqrt(x2 - xav * xav);
            ro /= Math.Sqrt(y2 - yav * yav);
            return ro;
        }

        static void ReadDataAirfoilSelfNoise()
        {
            string fileName = @"..\..\..\airfoil_self_noise.csv";
            if (!File.Exists(fileName))
            {
                Console.WriteLine("File not found {0}", fileName);
                Environment.Exit(0);
            }
            string[] lines = System.IO.File.ReadAllLines(fileName);
            for (int m = 1; m < lines.Length; ++m) //we skip first line
            {
                string[] sdata = lines[m].Split(';');
                if (6 != sdata.Length)
                {
                    Console.WriteLine("Misformatted data");
                    Environment.Exit(0);
                }
                double[] data = new double[5];
                for (int i = 0; i < data.Length; ++i)
                {
                    Double.TryParse(sdata[i], out data[i]);
                }
                double target = 0.0;
                Double.TryParse(sdata[5], out target);
                _inputs.Add(data);
                _target.Add(target);
            }

            _targetMin = _target.Min();
            _targetMax = _target.Max();
        }

        static void Main(string[] args)
        {
            ReadDataAirfoilSelfNoise();
            DateTime start = DateTime.Now;
            int M = 11;
            int NFolds = 16;
            double min_error = Double.MaxValue;
            double max_pearson = Double.MinValue;
            for (int count = 0; count < NFolds; ++count)
            {
                MakeTrainSelectValidateSets();
                GenerateInitialOperators(M, 16);
                BuildRepresentation(M, 3);
                double selectionError = RMSESelection();
                double pearson = ComputeCorrelationCoeff();
                if (selectionError < min_error)
                {
                    min_error = selectionError;
                    max_pearson = pearson;
                }
                Console.WriteLine("Run = {0}, selection sample error = {1:0.0000}, test sample pearson correlation coeff. = {2:0.0000}",
                    count, selectionError, pearson);
            }

            Console.WriteLine("\nWe select model with lowest selection error, Pearson for validation set is {0:0.0000}", max_pearson);
            DateTime end = DateTime.Now;
            TimeSpan duration = end - start;
            double time = duration.Minutes * 60.0 + duration.Seconds + duration.Milliseconds / 1000.0;
            Console.WriteLine("Total time for {0} training sessions {1:####.00} seconds", NFolds, time);
        }
    }
}
